"""Utility functions"""

from PIL import Image, ImageDraw

Lines = list[int]  # Lines are a list of pixel indices for an image
Mesh = tuple[
    Lines, Lines
]  # A mesh is a tuple of lists of x coordinates and y coordinates for lines


def crop_border(image: Image.Image, num_pixels: int = 1) -> Image.Image:
    """
    Crop the boder of an image by a few pixels.
    Sometimes when requesting an image from GPT-4o with a transparent background,
    the boarder pixels will not be transparent, so just remove them.
    """
    width, height = image.size
    box = (num_pixels, num_pixels, width - num_pixels, height - num_pixels)
    cropped = image.crop(box)
    return cropped


def overlay_grid_lines(
    image: Image.Image,
    mesh: Mesh,
    line_color: tuple[int, int, int] = (255, 0, 0),
    line_width: int = 1,
) -> Image.Image:
    """
    Overlay mesh which includes vertical (lines_x) and horizontal (lines_y) grid lines
    over image for visualization.
    """
    # Ensure we draw on an RGBA canvas
    canvas = image.convert("RGBA")
    draw = ImageDraw.Draw(canvas)

    lines_x, lines_y = mesh

    w, h = canvas.size
    # Draw each vertical line
    for x in lines_x:
        draw.line([(x, 0), (x, h)], fill=(*line_color, 255), width=line_width)

    # Draw each horizontal line
    for y in lines_y:
        draw.line([(0, y), (w, y)], fill=(*line_color, 255), width=line_width)

    return canvas


def scale_img(img: Image.Image, scale: int) -> Image.Image:
    """Scales the image up via nearest neightbor by scale factor."""
    w, h = img.size
    w_new, h_new = int(w * scale), int(h * scale)
    new_size = w_new, h_new
    scaled_img = img.resize(new_size, resample=Image.Resampling.NEAREST)
    return scaled_img



#!/usr/bin/env python3
"""Generate pixel art images using OpenAI's gpt-image-1.5 API and pixelate them."""

import argparse
import base64
import os
from datetime import datetime
from io import BytesIO
from pathlib import Path

from dotenv import load_dotenv
from openai import OpenAI
from PIL import Image

from proper_pixel_art.cli import add_pixelation_args
from proper_pixel_art.pixelate import pixelate


def load_api_key() -> str:
    """Load OpenAI API key from .env file."""
    load_dotenv()
    api_key = os.getenv("OPENAI_API_KEY")
    if not api_key:
        raise ValueError(
            "OPENAI_API_KEY not found in environment. "
            "Please create a .env file with your API key. "
            "See scripts/README.md for setup instructions."
        )
    return api_key


def parse_args() -> argparse.Namespace:
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(
        description="Generate images using OpenAI's gpt-image-1.5 and pixelate them.",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
Examples:
  # Basic usage
  python scripts/generate_pixel_art.py --prompt "A cute pixel art cat"

  # With all OpenAI options
  python scripts/generate_pixel_art.py \\
    --prompt "A 16 bit pixel art fantasy castle" \\
    --size 1792x1024 \\
    --n 2

  # With pixelation options
  python scripts/generate_pixel_art.py \\
    --prompt "A 16 bit pixel art robot" \\
    --colors 16 \\
    --scale-result 20 \\
    --transparent
        """,
    )

    # OpenAI API parameters
    openai_group = parser.add_argument_group("OpenAI API options")
    openai_group.add_argument(
        "--prompt",
        type=str,
        required=True,
        help="Text description for image generation.",
    )
    openai_group.add_argument(
        "--size",
        type=str,
        choices=["1024x1024", "1024x1536", "1536x1024", "auto"],
        default="1024x1024",
        help="Generated image size (default: 1024x1024).",
    )
    openai_group.add_argument(
        "--n",
        type=int,
        choices=range(1, 11),
        default=1,
        help="Number of images to generate (1-10, default: 1).",
    )

    # Add common pixelation arguments
    add_pixelation_args(parser)

    # Output parameters
    parser.add_argument(
        "-o",
        "--output-dir",
        dest="output_dir",
        type=Path,
        default=Path("."),
        help="Output directory for generated images (default: current directory).",
    )

    return parser.parse_args()


def generate_images(client: OpenAI, args: argparse.Namespace) -> list[Image.Image]:
    """Generate images using OpenAI API."""
    print(f"Generating {args.n} image(s) with prompt: '{args.prompt}'")
    print(f"Settings: size={args.size}")

    response = client.images.generate(
        model="gpt-image-1.5",
        prompt=args.prompt,
        n=args.n,
        size=args.size,
    )

    # Convert base64 images to PIL Images
    images = []
    for image_data in response.data:
        image_bytes = base64.b64decode(image_data.b64_json)
        images.append(Image.open(BytesIO(image_bytes)))

    return images


def process_image(
    original_image: Image.Image,
    args: argparse.Namespace,
    timestamp: str,
    index: int = 0,
) -> tuple[Path, Path]:
    """Pixelate and save a single image."""
    print(f"Processing image {index + 1}...")

    # Generate filenames
    suffix = f"_{index}" if args.n > 1 else ""
    original_filename = f"{timestamp}{suffix}_original.png"
    pixelated_filename = f"{timestamp}{suffix}_pixelated.png"

    original_path = args.output_dir / original_filename
    pixelated_path = args.output_dir / pixelated_filename

    # Save original
    original_image.save(original_path)
    print(f"Saved original: {original_path}")

    # Pixelate
    print(f"Pixelating image {index + 1}...")
    pixelated_image = pixelate(
        original_image,
        num_colors=args.num_colors,
        initial_upscale_factor=args.initial_upscale,
        scale_result=args.scale_result,
        transparent_background=args.transparent,
        pixel_width=args.pixel_width,
    )

    # Save pixelated
    pixelated_image.save(pixelated_path)
    print(f"Saved pixelated: {pixelated_path}")

    return original_path, pixelated_path


def main() -> None:
    """Main entry point."""
    args = parse_args()

    # Create output directory
    args.output_dir.mkdir(parents=True, exist_ok=True)

    # Load API key and create client
    try:
        api_key = load_api_key()
        client = OpenAI(api_key=api_key)
    except ValueError as e:
        print(f"Error: {e}")
        return

    # Generate timestamp for this batch
    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")

    try:
        # Generate images via OpenAI API
        images = generate_images(client, args)

        # Process each image
        for i, img in enumerate(images):
            process_image(img, args, timestamp, i)

        print(f"\nSuccessfully processed {len(images)} image(s)")
        print(f"Output directory: {args.output_dir.absolute()}")

    except Exception as e:
        print(f"Error occurred: {type(e).__name__}: {e}")
        raise


if __name__ == "__main__":
    main()



from pathlib import Path

import pytest


@pytest.fixture(name="assets")
def fixture_assets() -> Path:
    assets = Path.cwd() / "assets"
    return assets


@pytest.fixture(name="pixelate_png_test_params")
def fixture_pixelate_png_test_params(assets: Path) -> dict[str, dict]:
    pixelate_png_test_params = {
        # Test case for transparent background with hole
        "anchor": {
            "num_colors": 16,
            "result_scale": 5,
            "transparent_background": True,
            "path": assets / "anchor" / "anchor.png",
        },
        "ash": {
            "num_colors": 16,
            "result_scale": 5,
            "transparent_background": False,
            "path": assets / "ash" / "ash.png",
        },
        "bat": {
            "num_colors": 16,
            "result_scale": 5,
            "transparent_background": True,
            "path": assets / "bat" / "bat.png",
        },
        "blob": {
            "num_colors": 16,
            "result_scale": 25,
            "transparent_background": False,
            "path": assets / "blob" / "blob.png",
        },
        "demon": {
            "num_colors": 64,
            "result_scale": 5,
            "transparent_background": True,
            "path": assets / "demon" / "demon.png",
        },
        "mountain": {
            "num_colors": 64,
            "result_scale": 5,
            "transparent_background": False,
            "path": assets / "mountain" / "mountain.png",
        },
        # Test case for skipping quantization
        "pumpkin": {
            "num_colors": None,
            "result_scale": 5,
            "transparent_background": False,
            "path": assets / "pumpkin" / "pumpkin.png",
        },
    }
    return pixelate_png_test_params



"""Handles image colors logic"""

from collections import Counter
from pathlib import Path

import numpy as np
from PIL import Image, ImageColor
from PIL.Image import Quantize

RGB = tuple[int, int, int]
RGBA = tuple[int, int, int, int]

# Alpha threshold for determining pixel opacity
# Pixels with alpha >= this value are considered opaque
ALPHA_THRESHOLD = 128


def _is_majority_transparent(opaque_count: int, total_count: int) -> bool:
    return opaque_count <= total_count / 2


def _rgb_dist(a: RGB, b: RGB) -> int:
    """Naive color distance"""
    dr, dg, db = a[0] - b[0], a[1] - b[1], a[2] - b[2]
    return dr**2 + dg**2 + db**2


def _top_opaque_colors(
    img: Image.Image, alpha_threshold: int, limit: int = 8
) -> list[RGB]:
    """Return the most common opaque colors (RGB) up to limit."""
    rgba = img.convert("RGBA").copy()
    rgba.thumbnail((160, 160))  # speed and de-noise tiny details
    counts = Counter()
    for r, g, b, a in rgba.getdata():
        if a >= alpha_threshold:
            counts[(r, g, b)] += 1
    return [c for c, _ in counts.most_common(limit)]


def _pick_background(colors: list[RGB]) -> RGB:
    """
    Pick the candidate farthest from the common colors.
    Used for choosing the color of pixels with alpha to avoid clashing
    with the actual colors in the image
    """
    background_color_candidates: list[RGB] = [
        (0, 255, 255),  # cyan
        (255, 255, 255),  # white
        (255, 0, 0),  # red
        (0, 255, 0),  # green
        (0, 0, 255),  # blue
        (255, 255, 0),  # yellow
        (255, 0, 255),  # magenta
        (255, 128, 0),  # orange
        (128, 0, 255),  # violet
        (0, 128, 255),  # sky
        (0, 255, 128),  # mint
        (255, 0, 128),  # pink
    ]

    if not colors:
        return (255, 255, 255)
    best, best_score = background_color_candidates[0], -1
    for color_candidate in background_color_candidates:
        score = min(_rgb_dist(color_candidate, c) for c in colors)
        if score > best_score:
            best, best_score = color_candidate, score
    return best


def clamp_alpha(
    image: Image.Image,
    alpha_threshold: int = ALPHA_THRESHOLD,
    mode: str = "RGB",
    background_hex: str | None = None,
) -> Image.Image:
    """
    Replace pixels with alpha < threshold by a background color.
    If background_hex is None, choose a color far from the most common image colors.
    mode: 'RGB' or 'L'
    """
    if mode not in ("RGB", "L"):
        raise ValueError("mode must be 'RGB' or 'L'")

    if background_hex is None:
        common = _top_opaque_colors(image, alpha_threshold)
        bg_rgb = _pick_background(common)
    else:
        bg_rgb = ImageColor.getrgb(background_hex)

    base = image.convert(mode)
    alpha = image.getchannel("A")
    mask = alpha.point(lambda p: 255 if p >= alpha_threshold else 0)

    background = Image.new("RGB", image.size, bg_rgb).convert(mode)

    return Image.composite(base, background, mask)


def extract_and_scale_alpha(image: Image.Image, scale_factor: int = 1) -> np.ndarray:
    """
    Extract alpha channel from an RGBA image and scale it.

    Args:
        image: RGBA PIL Image
        scale_factor: Integer scale factor for resizing (default 1 = no scaling)

    Returns:
        Numpy array of alpha channel values (uint8), scaled if scale_factor != 1
    """
    # Extract alpha channel from RGBA image
    alpha_channel = np.array(image.convert("RGBA"))[:, :, 3]

    # Scale alpha channel if needed
    if scale_factor != 1:
        alpha_img = Image.fromarray(alpha_channel, mode="L")
        new_size = (alpha_img.width * scale_factor, alpha_img.height * scale_factor)
        scaled_alpha_img = alpha_img.resize(new_size, Image.Resampling.NEAREST)
        return np.array(scaled_alpha_img)
    else:
        return alpha_channel


def get_opaque_cell_color(cell_pixels: np.ndarray) -> RGBA:
    """
    cell_pixels: shape (height_cell, width_cell, 3), dtype=uint8
    returns the most frequent RGB tuple in the cell_pixels block, with 255 in fourth entry for opaque alpha.
    """
    # flatten to tuple of pixel values
    flat = list(map(tuple, cell_pixels.reshape(-1, 3)))
    cell_color = Counter(flat).most_common(1)[0][0]
    return (*cell_color, 255)


def get_cell_color_with_alpha(cell_pixels: np.ndarray, cell_alpha: np.ndarray) -> RGBA:
    """
    Select representative color for a quantized cell, considering transparency.

    Decision logic:
    - If >=50% of pixels are transparent (alpha < ALPHA_THRESHOLD), return (0,0,0,0)
    - Otherwise, return the most common RGB color with full opacity (R,G,B,255)

    Args:
        cell_pixels: shape (height, width, 3), dtype=uint8 (RGB from quantized image)
        cell_alpha: shape (height, width), dtype=uint8 (alpha channel from original)

    Returns:
        RGBA tuple - either (0,0,0,0) for transparent or (R,G,B,255) for opaque
    """
    total_pixels = cell_alpha.size
    opaque_count = np.sum(cell_alpha >= ALPHA_THRESHOLD)

    # If majority is transparent (>= 50%), return fully transparent
    if _is_majority_transparent(opaque_count, total_pixels):
        return (0, 0, 0, 0)

    # Otherwise return most common color with full opacity
    cell_color = get_opaque_cell_color(cell_pixels)
    return cell_color


def _dominant_rgb_by_binning(rgb_pixels: np.ndarray) -> RGB:
    """
    Find the dominant color using offset binning algorithm.

    Algorithm (offset binning):
    1. Quantize RGB space into bins using two offset grids
       - Grid 1: boundaries at 0, 52, 104, 156, 208
       - Grid 2: boundaries at 26, 78, 130, 182, 234 (offset by half bin size)
    2. For each grid, find the bin with the most pixels
    3. Use the grid that produces the larger dominant cluster
    4. Return the median color of pixels in that bin

    This avoids bin boundary artifacts: colors straddling one grid's boundary
    will be grouped together in the other grid.

    For small pixel counts (<=3), the binning algorithm is skipped and
    either the single pixel value or median is returned directly.

    Args:
        rgb_pixels: shape (N, 3), dtype=uint8 - flat array of RGB values

    Returns:
        RGB tuple of the representative color
    """
    # Edge cases for small pixel counts - binning doesn't make sense
    if len(rgb_pixels) == 1:
        r, g, b = rgb_pixels[0]
        return (int(r), int(g), int(b))
    if len(rgb_pixels) <= 3:
        median = np.median(rgb_pixels, axis=0).astype(np.uint8)
        return (int(median[0]), int(median[1]), int(median[2]))

    bin_size = 52

    # Grid 1: standard binning (boundaries at 0, 52, 104...)
    bins1 = rgb_pixels // bin_size
    indices1 = bins1[:, 0] * 25 + bins1[:, 1] * 5 + bins1[:, 2]
    counts1 = np.bincount(indices1, minlength=125)
    dominant1 = np.argmax(counts1)
    max_count1 = counts1[dominant1]

    # Grid 2: offset binning (boundaries at 26, 78, 130...)
    # Add offset before dividing, clamp to avoid overflow
    offset = bin_size // 2
    bins2 = np.minimum(rgb_pixels + offset, 255) // bin_size
    indices2 = bins2[:, 0] * 25 + bins2[:, 1] * 5 + bins2[:, 2]
    counts2 = np.bincount(indices2, minlength=125)
    dominant2 = np.argmax(counts2)
    max_count2 = counts2[dominant2]

    # Use the grid with the larger dominant cluster
    if max_count1 >= max_count2:
        mask = indices1 == dominant1
    else:
        mask = indices2 == dominant2

    dominant_pixels = rgb_pixels[mask]

    # Return median of dominant bin (robust to outliers within bin)
    return tuple(np.median(dominant_pixels, axis=0).astype(np.uint8))


def get_cell_color_skip_quantization(
    cell_pixels: np.ndarray, alpha_threshold: int = ALPHA_THRESHOLD
) -> RGBA:
    """
    Select representative RGBA color for a cell when quantization is skipped.

    Used when quantization is skipped to preserve original colors
    while handling noise/grain and filtering out background bleed-in.

    Decision logic:
    - If >=50% of pixels are transparent (alpha < threshold), return fully transparent (0,0,0,0)
    - Otherwise, return fully opaque color computed from only the opaque pixels

    The color is determined by finding the densest color region using offset binning,
    which avoids bin boundary artifacts: colors straddling one grid's boundary
    will be grouped together in the other grid.

    Args:
        cell_pixels: shape (height, width, 4), dtype=uint8 (RGBA)
        alpha_threshold: minimum alpha to consider pixel opaque (default ALPHA_THRESHOLD)

    Returns:
        RGBA tuple - either (0,0,0,0) for transparent or (R,G,B,255) for opaque
    """
    pixels = cell_pixels.reshape(-1, 4)
    total_pixels = len(pixels)

    # Edge case: empty cell
    if total_pixels == 0:
        return (0, 0, 0, 0)

    # Filter to opaque pixels only
    opaque_mask = pixels[:, 3] >= alpha_threshold
    opaque_pixels = pixels[opaque_mask]

    # If majority is transparent (>= 50%), return fully transparent
    if _is_majority_transparent(len(opaque_pixels), total_pixels):
        return (0, 0, 0, 0)

    # Get RGB of opaque pixels and find dominant color
    rgb_pixels = opaque_pixels[:, :3]
    r, g, b = _dominant_rgb_by_binning(rgb_pixels)
    return (int(r), int(g), int(b), 255)


def palette_img(
    image: Image.Image,
    num_colors: int = 16,
    quantize_method: int = Quantize.MAXCOVERAGE,
    output_dir: Path | None = None,
) -> Image.Image:
    """
    Discretizes the colors in the image to at most num_colors.
    Saves the quantized image to output_dir if not None.
    Returns the color pallete of the image.

    The maximum coverage algorithm is used by default as the quantization method.
    Emperically this algorithm proivdes the best results overall, although
    for some examples num_colors needs to be chosen very large even when the
    image has a small number of actual colors. In these instances, Quantize.FASTOCTREE
    can work instead.

    If the colors of the result don't look right, try increasing num_colors.
    """
    image_rgb = clamp_alpha(image, mode="RGB")
    quantized_img = image_rgb.quantize(
        colors=num_colors, method=quantize_method, dither=Image.Dither.NONE
    )
    if output_dir is not None:
        quantized_img.save(output_dir / "quantized_original.png")
    return quantized_img


def most_common_boundary_color(image: Image.Image) -> RGB:
    """
    Return the exact RGB color that occurs most on the image boundary.
    """
    image_rgb = image.convert("RGB")
    w, h = image_rgb.size

    # top and bottom rows
    top = list(image_rgb.crop((0, 0, w, 1)).getdata())
    bottom = list(image_rgb.crop((0, h - 1, w, h)).getdata())
    # left and right columns (excluding corners)
    left = [image_rgb.getpixel((0, y)) for y in range(1, h - 1)]
    right = [image_rgb.getpixel((w - 1, y)) for y in range(1, h - 1)]

    counts = Counter(top + bottom + left + right)
    mode_color = counts.most_common(1)[0][0]
    return mode_color  # (R, G, B)


def make_background_transparent(image: Image.Image) -> Image.Image:
    """
    Make the background fully transparent by:
      1) Identifying the most common color on the image boundary
      2) Setting alpha=0 for all pixels matching that color

    Note: This sets transparency for ALL pixels matching the boundary color,
    not just boundary pixels (no flood fill).
    """
    background_color = most_common_boundary_color(image)
    image_rgba = image.convert("RGBA")
    px = list(image_rgba.getdata())

    out = []
    for r, g, b, a in px:
        # If the color is the same as the background color, make it transparent
        if (r, g, b) == background_color:
            out.append((r, g, b, 0))
        else:
            out.append((r, g, b, a or 255))

    image_rgba.putdata(out)
    return image_rgba


def main():
    img_path = Path.cwd() / "assets" / "blob" / "blob.png"
    img = Image.open(img_path).convert("RGBA")
    paletted = palette_img(img)
    paletted.show()


if __name__ == "__main__":
    main()



"""Visual output tests"""

from pathlib import Path

from PIL import Image

from proper_pixel_art import pixelate


def test_pixelate_pngs(pixelate_png_test_params: dict[str, dict]) -> None:
    """Test that pixelate algorithm runs without error for png images and generates output."""
    output_dir = Path.cwd() / "tests" / "outputs"
    output_dir.mkdir(exist_ok=True, parents=True)
    for name, params in pixelate_png_test_params.items():
        pixel_dir = output_dir / str(name)
        pixel_dir.mkdir(parents=True, exist_ok=True)

        img = Image.open(params["path"])

        # Pixelate
        result = pixelate.pixelate(
            img,
            num_colors=params["num_colors"],
            scale_result=params["result_scale"],
            transparent_background=params["transparent_background"],
            intermediate_dir=pixel_dir,
        )

        # Save the result
        result_path = pixel_dir / "result.png"
        result.save(result_path)

        # Verify the output exists and has a width and height
        assert result_path.exists(), f"Output file not created for {name}"
        assert result.width > 0 and result.height > 0, f"Invalid dimensions for {name}"

        print(f"Generated output for {name}: {result_path}")
    print(
        (
            "Successfully generated all .png test images.\n"
            f"Manually inspect the results in {output_dir} to verify pixelation quality."
        )
    )



"""Tests for the colors module."""

import numpy as np

from proper_pixel_art import colors


def _make_rgba(rgb_cell: np.ndarray, alpha: int = 255) -> np.ndarray:
    """Helper to convert RGB cell to RGBA with given alpha."""
    h, w = rgb_cell.shape[:2]
    rgba = np.zeros((h, w, 4), dtype=np.uint8)
    rgba[:, :, :3] = rgb_cell
    rgba[:, :, 3] = alpha
    return rgba


class TestGetCellColor:
    """Tests for the mode-based color selection (for quantized images)."""

    def test_most_common_color_selected(self):
        """Returns the most frequent color in the cell."""
        cell = np.zeros((10, 10, 3), dtype=np.uint8)
        cell[:6, :] = [255, 0, 0]  # 60 red pixels
        cell[6:, :] = [0, 0, 255]  # 40 blue pixels

        result = colors.get_opaque_cell_color(cell)
        assert result == (255, 0, 0, 255)

    def test_single_color(self):
        """Single color cell returns that color."""
        cell = np.full((5, 5, 3), [42, 84, 126], dtype=np.uint8)
        result = colors.get_opaque_cell_color(cell)
        assert result == (42, 84, 126, 255)


class TestGetCellColorWithAlpha:
    """Tests for get_cell_color_with_alpha (quantized RGB + alpha channel)."""

    def test_majority_transparent_returns_transparent(self):
        """Cell with >=50% transparent pixels returns fully transparent."""
        # RGB cell from quantized image
        cell_pixels = np.full((10, 10, 3), [200, 100, 50], dtype=np.uint8)

        # Alpha channel from original image - 70% transparent
        cell_alpha = np.zeros((10, 10), dtype=np.uint8)
        cell_alpha[:3, :] = 255  # 30% opaque (30 pixels)
        # Remaining 70% stays at 0 (transparent)

        result = colors.get_cell_color_with_alpha(cell_pixels, cell_alpha)
        assert result == (0, 0, 0, 0)

    def test_majority_opaque_returns_most_common_color(self):
        """Cell with >50% opaque pixels returns the most common RGB color."""
        # Quantized RGB cell with multiple colors
        cell_pixels = np.zeros((10, 10, 3), dtype=np.uint8)
        cell_pixels[:7, :] = [255, 0, 0]  # 70% red
        cell_pixels[7:, :] = [0, 0, 255]  # 30% blue

        # Alpha channel from original - 70% opaque
        cell_alpha = np.zeros((10, 10), dtype=np.uint8)
        cell_alpha[:7, :] = 255  # 70% opaque
        # Remaining 30% transparent

        result = colors.get_cell_color_with_alpha(cell_pixels, cell_alpha)
        assert result == (255, 0, 0, 255)  # Most common color (red) with full opacity

    def test_exactly_50_percent_transparent_returns_transparent(self):
        """Cell with exactly 50% transparent (tie) returns transparent."""
        cell_pixels = np.full((10, 10, 3), [100, 100, 100], dtype=np.uint8)

        # Exactly 50% opaque, 50% transparent
        cell_alpha = np.zeros((10, 10), dtype=np.uint8)
        cell_alpha[:5, :] = 255  # 50% opaque
        # Remaining 50% transparent

        result = colors.get_cell_color_with_alpha(cell_pixels, cell_alpha)
        assert result == (0, 0, 0, 0)

    def test_single_color_all_opaque(self):
        """Single color cell with all opaque pixels returns that color."""
        cell_pixels = np.full((5, 5, 3), [42, 84, 126], dtype=np.uint8)
        cell_alpha = np.full((5, 5), 255, dtype=np.uint8)  # All opaque

        result = colors.get_cell_color_with_alpha(cell_pixels, cell_alpha)
        assert result == (42, 84, 126, 255)

    def test_most_common_color_selected_from_quantized(self):
        """Returns the most frequent color from quantized RGB cell."""
        # Quantized cell with clear most-common color
        cell_pixels = np.zeros((10, 10, 3), dtype=np.uint8)
        cell_pixels[:8, :] = [200, 50, 25]  # 80% this color
        cell_pixels[8:, :] = [100, 150, 75]  # 20% this color

        # All pixels opaque
        cell_alpha = np.full((10, 10), 255, dtype=np.uint8)

        result = colors.get_cell_color_with_alpha(cell_pixels, cell_alpha)
        assert result == (200, 50, 25, 255)


class TestGetCellColorSkipQuantization:
    """Tests for histogram-based color selection (when quantization is skipped).

    The function now takes RGBA input and returns RGBA output.
    For fully opaque cells (alpha=255), the RGB result matches the original algorithm.
    """

    def test_single_color_returns_that_color(self):
        """Cell with single color returns that color."""
        rgb = np.full((10, 10, 3), [255, 0, 0], dtype=np.uint8)
        cell = _make_rgba(rgb)
        result = colors.get_cell_color_skip_quantization(cell)
        assert result == (255, 0, 0, 255)

    def test_uniform_with_outliers_filters_outliers(self):
        """Mostly uniform cell with few outliers returns the uniform color, not average."""
        rgb = np.full((10, 10, 3), [200, 100, 50], dtype=np.uint8)
        # Add a few outlier pixels (background bleed-in) - very different colors
        rgb[0, 0] = [0, 0, 255]  # Blue outlier - different bin
        rgb[0, 1] = [0, 255, 0]  # Green outlier - different bin
        cell = _make_rgba(rgb)

        result = colors.get_cell_color_skip_quantization(cell)
        r, g, b, a = result
        # Should be close to the dominant color (200, 100, 50), not skewed by outliers
        assert 190 <= r <= 210, f"Expected r near 200, got {r}"
        assert 90 <= g <= 110, f"Expected g near 100, got {g}"
        assert 40 <= b <= 60, f"Expected b near 50, got {b}"
        assert a == 255

    def test_two_color_groups_returns_dominant(self):
        """Cell with two distinct color groups returns the dominant one."""
        rgb = np.zeros((10, 10, 3), dtype=np.uint8)
        rgb[:7, :] = [255, 0, 0]  # 70% red (bin 7,0,0)
        rgb[7:, :] = [0, 0, 255]  # 30% blue (bin 0,0,7)
        cell = _make_rgba(rgb)

        result = colors.get_cell_color_skip_quantization(cell)
        r, g, b, a = result
        # Should be close to red (the dominant color)
        assert r > 200, f"Expected red-dominant result, got {result}"
        assert b < 50, f"Expected blue filtered out, got {result}"
        assert a == 255

    def test_noisy_image_returns_stable_color(self):
        """Noisy/grainy image returns a stable average, not erratic pixel."""
        rgb = np.zeros((10, 10, 3), dtype=np.uint8)
        # Simulate film grain: many unique red-ish pixels (all in same bin ~250)
        for i in range(7):
            for j in range(10):
                rgb[i, j] = [250 + (i * j % 6), (i + j) % 5, (i - j) % 5]
        # Some blue-ish pixels with noise (minority, different bin)
        for i in range(7, 10):
            for j in range(10):
                rgb[i, j] = [(i + j) % 5, (i - j) % 5, 250 + (i * j % 6)]
        cell = _make_rgba(rgb)

        result = colors.get_cell_color_skip_quantization(cell)
        r, g, b, a = result
        # Result should be red (the dominant color bin)
        assert r > 200, f"Expected red dominant, got {result}"
        assert a == 255

    def test_gradient_returns_value_in_range(self):
        """Cell with gradient returns a value within the gradient range."""
        rgb = np.zeros((10, 10, 3), dtype=np.uint8)
        # Create a red gradient from 50 to 200
        for i in range(10):
            for j in range(10):
                rgb[i, j, 0] = 50 + i * 15 + j  # Red varies ~50-200
        cell = _make_rgba(rgb)

        result = colors.get_cell_color_skip_quantization(cell)
        r, g, b, a = result
        # Result should be somewhere in the gradient range (mode bin + neighbors)
        assert 50 <= r <= 200, f"Red channel out of expected range, got {result}"
        assert a == 255

    def test_small_cell_still_works(self):
        """Very small cells (like 2x2) still return valid results."""
        rgb = np.array(
            [[[255, 0, 0], [255, 0, 0]], [[0, 0, 255], [0, 255, 0]]], dtype=np.uint8
        )
        cell = _make_rgba(rgb)
        result = colors.get_cell_color_skip_quantization(cell)
        assert len(result) == 4
        assert all(0 <= c <= 255 for c in result)
        assert result[3] == 255  # Alpha should be opaque

    def test_single_pixel_cell(self):
        """Single pixel cell returns that pixel's color."""
        rgb = np.array([[[128, 64, 32]]], dtype=np.uint8)
        cell = _make_rgba(rgb)
        result = colors.get_cell_color_skip_quantization(cell)
        assert result == (128, 64, 32, 255)

    def test_empty_cell(self):
        """Empty cell (0 pixels) returns transparent."""
        cell = np.zeros((0, 0, 4), dtype=np.uint8)
        result = colors.get_cell_color_skip_quantization(cell)
        assert result == (0, 0, 0, 0)

    def test_bin_boundary_handled(self):
        """Colors at bin boundary (e.g., 32) are handled via neighbor merging."""
        rgb = np.zeros((10, 10, 3), dtype=np.uint8)
        # Values spanning boundary at 32 (bin 0/1 boundary)
        for i in range(10):
            for j in range(10):
                # Values 27-37, spanning the boundary at 32
                val = 27 + (i + j) % 11
                rgb[i, j] = [val, val, val]
        cell = _make_rgba(rgb)

        result = colors.get_cell_color_skip_quantization(cell)
        r, g, b, a = result
        # Should include all pixels via neighbor merging, average ~32
        assert 27 <= r <= 37, f"Expected value near 32, got {r}"
        assert r == g == b, f"Expected uniform gray, got {result}"
        assert a == 255

    def test_dark_to_green_gradient_returns_green_ish(self):
        """
        Gradient from dark (black) to bright green should return
        a noticeably green color, not black-ish.

        This tests the case where dark pixels dominate by count
        but green is the distinctive color.
        """
        rgb = np.zeros((10, 10, 3), dtype=np.uint8)
        # Create gradient: columns go from black to bright green
        for i in range(10):
            for j in range(10):
                # R: stays low
                # G: increases from ~20 to ~150
                # B: stays low
                intensity = i * 10 + j  # 0-99, scaled
                rgb[i, j] = [
                    20 + intensity // 10,  # R: 20-29
                    20 + intensity,  # G: 20-119
                    20 + intensity // 10,  # B: 20-29
                ]
        cell = _make_rgba(rgb)

        result = colors.get_cell_color_skip_quantization(cell)
        r, g, b, a = result
        # Green should be meaningfully higher than red/blue
        # (showing green tint, not just dark gray)
        assert g > r + 10, f"Expected green tint, got {result}"
        assert g > b + 10, f"Expected green tint, got {result}"
        assert a == 255

    def test_majority_transparent_returns_transparent(self):
        """Cell with >=50% transparent pixels returns fully transparent."""
        cell = np.zeros((10, 10, 4), dtype=np.uint8)
        cell[:4, :, :3] = [100, 100, 100]  # 40% opaque
        cell[:4, :, 3] = 255
        cell[4:, :, 3] = 0  # 60% transparent

        result = colors.get_cell_color_skip_quantization(cell)
        assert result == (0, 0, 0, 0)

    def test_majority_opaque_returns_opaque_color(self):
        """Cell with >50% opaque pixels returns the opaque color."""
        cell = np.zeros((10, 10, 4), dtype=np.uint8)
        cell[:6, :, :3] = [100, 100, 100]  # 60% opaque
        cell[:6, :, 3] = 255
        cell[6:, :, 3] = 0  # 40% transparent

        result = colors.get_cell_color_skip_quantization(cell)
        r, g, b, a = result
        assert a == 255
        assert 90 <= r <= 110

    def test_exactly_50_percent_transparent_returns_transparent(self):
        """Cell with exactly 50% transparent (tie) returns transparent."""
        cell = np.zeros((10, 10, 4), dtype=np.uint8)
        cell[:5, :, :3] = [100, 100, 100]  # 50% opaque
        cell[:5, :, 3] = 255
        cell[5:, :, 3] = 0  # 50% transparent

        result = colors.get_cell_color_skip_quantization(cell)
        assert result == (0, 0, 0, 0)



"""Handles mesh detection from pixel art style images"""

from pathlib import Path

import cv2
import numpy as np
from PIL import Image

from proper_pixel_art import colors, utils
from proper_pixel_art.utils import Lines, Mesh


def close_edges(edges: np.ndarray, kernel_size: int = 10) -> np.ndarray:
    """
    Apply a morphological closing to fill small gaps in edge map.
    """
    # Use a rectangular kernel of size kernel_size x kernel_size
    kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (kernel_size, kernel_size))
    closed = cv2.morphologyEx(edges, cv2.MORPH_CLOSE, kernel)
    return closed


def cluster_lines(lines: Lines, threshold: int = 4) -> Lines:
    """Remove lines that are too close to each other by clustering near values"""
    if not lines:
        return []
    lines = sorted(lines)
    clusters = [[lines[0]]]
    for p in lines[1:]:
        if abs(p - clusters[-1][-1]) <= threshold:
            clusters[-1].append(p)
        else:
            clusters.append([p])
    # use the median of each cluster
    return [int(np.median(cluster)) for cluster in clusters]


def detect_grid_lines(
    edges: np.ndarray,
    hough_rho: float = 1.0,
    hough_theta_rad: float = np.deg2rad(1),
    hough_threshold: int = 100,
    hough_min_line_len: int = 50,
    hough_max_line_gap: int = 10,
    angle_threshold_deg=15,
) -> Mesh:
    """
    - Use Hough line transformation to detect the pixel edges.
    - Only keep lines that are close to vertical or horizontal
    - Cluster the lines so they aren't too close
    Return:
    - two lists: x-coordinates (vertical lines) and y-coordinates (horizontal lines)
    """
    hough_lines = cv2.HoughLinesP(
        edges,
        hough_rho,
        hough_theta_rad,
        hough_threshold,
        minLineLength=hough_min_line_len,
        maxLineGap=hough_max_line_gap,
    )

    height, width = edges.shape
    # Include the sides of the image in lines since they aren't detected by the Hough transform
    lines_x, lines_y = [0, width - 1], [0, height - 1]
    if hough_lines is None:
        return lines_x, lines_y

    # Loop over all detected lines, only keep the ones that are close to vertical or horizontal
    for x1, y1, x2, y2 in hough_lines[:, 0]:
        dx, dy = x2 - x1, y2 - y1
        angle = abs(np.arctan2(dy, dx))
        # vertical if angle > 90-threshold, horizontal if angle < threshold
        if angle > np.deg2rad(90 - angle_threshold_deg):
            lines_x.append(round((x1 + x2) / 2))
        elif angle < np.deg2rad(angle_threshold_deg):
            lines_y.append(round((y1 + y2) / 2))

    # Finally cluster the lines so they aren't too close to each other
    clustered_lines_x = cluster_lines(lines_x)
    clustered_lines_y = cluster_lines(lines_y)
    return clustered_lines_x, clustered_lines_y


def get_pixel_width(
    line_collection: list[Lines], trim_outlier_fraction: float = 0.2
) -> int:
    """
    Takes list of line coordinates, and outlier fraction.
    Returns the predicted pixel width by filtering outliers and taking the median.
    We assume that the grid spacing accress all sets of lines,
    then all grid spacings are concatenated.

    The resulting width does not have to be perfect because the color of the pixels
    are detemined by which color is mostly in the corresponding cells.

    This method could be generalized to cases when the pixel size in the x direction
    is different from the y direction, then the width of each direction
    would have to be calculated separately.
    """
    all_gaps = []
    for lines in line_collection:
        gap = np.diff(lines)
        all_gaps.append(gap)
    gaps = np.concatenate(all_gaps)

    # Filter lower and upper percentile
    low = np.percentile(gaps, 100 * trim_outlier_fraction)
    hi = np.percentile(gaps, 100 * (1 - trim_outlier_fraction))
    middle = gaps[(gaps >= low) & (gaps <= hi)]
    if len(middle) == 0:
        # fallback to median of all gaps
        middle = gaps

    return np.median(middle)


def homogenize_lines(lines: Lines, pixel_width: int) -> Lines:
    """
    Given sorted line coords and pixel width,
    further partition those line coordinates to approximately even spacing.
    """
    section_widths = np.diff(lines)
    complete_lines = lines[:-1]
    for index, section_width in enumerate(section_widths):
        # Get number of pixels to partition section width into
        num_pixels = int(np.round(section_width / pixel_width))
        if num_pixels == 0:
            section_pixel_width = 0
        else:
            section_pixel_width = section_width / num_pixels
        line_start = lines[index]
        section_lines = [
            line_start + int(n * section_pixel_width) for n in range(num_pixels)
        ]
        # Replace the start index in completed lines with list of new line coordinates
        # Everything will be unpacked after to maintain indexes
        complete_lines[index] = section_lines

    complete_lines = [line for sublist in complete_lines for line in sublist]
    # Add last line back in because it was excluded earlier
    complete_lines.append(lines[-1])

    return complete_lines


def compute_mesh(
    img: Image.Image,
    canny_thresholds: tuple[int] = (50, 200),
    closure_kernel_size: int = 8,
    output_dir: Path | None = None,
    pixel_width: int | None = None,
) -> Mesh:
    """
    Finds grid lines of a high resolution noisy image.
    - Uses Canny edge detector to find vertical and horizontal edges
    - Closes small gaps between edges with morphological closing
    - Uses Hough transform to detect pixel edge lines
    - Finds true width of pixels from line differences
    - Completes mesh by filling in gaps between identified lines
    inputs:
        img: The image to compute the mesh
        canny_thresholds: thresholds 1 and 2 for canny edge detection algorithm
        closure_kernel_size: Kernel size for the morphological closure
        output_dir (optional): If set, saves images of steps in algorithm to dir

    output:
        Returns The pixel mesh: mesh_x, mesh_y
            tuple of two lists of integer coordinates ():
        - mesh_x: Coordinates of pixel mesh on the x-axis
        - mesh_y: Coordinates of pixel mesh on the y-axis

    Note: this could even be generalized to detect grid lines that
    have been distorted via linear transformation.
    """
    # Crop border and zero out mostly transparent pixels from alpha
    cropped_img = utils.crop_border(img, num_pixels=2)
    grey_img = colors.clamp_alpha(cropped_img, mode="L")

    # Find edges using Canny edge detection
    edges = cv2.Canny(np.array(grey_img), *canny_thresholds)

    # Close small gaps in edges with morphological closing
    closed_edges = close_edges(edges, kernel_size=closure_kernel_size)

    # Use Hough transform to get an initial estimate for pixel lines
    mesh_initial = detect_grid_lines(closed_edges)

    if pixel_width is None:
        # Get the true width of the pixels if a value hasn't been provided
        pixel_width = get_pixel_width(mesh_initial)

    # Fill in the gaps between the lines to complete the grid
    lines_x, lines_y = mesh_initial
    mesh_x = homogenize_lines(lines_x, pixel_width)
    mesh_y = homogenize_lines(lines_y, pixel_width)
    mesh_final = mesh_x, mesh_y

    if output_dir is not None:
        edges_img = Image.fromarray(edges, mode="L")
        edges_img.save(output_dir / "edges.png")
        closed_edges_img = Image.fromarray(closed_edges, mode="L")
        closed_edges_img.save(output_dir / "closed_edges.png")

        img_with_lines = utils.overlay_grid_lines(img, mesh_initial)
        img_with_lines.save(output_dir / "lines.png")
        img_with_completed_lines = utils.overlay_grid_lines(img, mesh_final)
        img_with_completed_lines.save(output_dir / "mesh.png")

    return mesh_final


def compute_mesh_with_scaling(
    img: Image.Image,
    upscale_factor: int,
    output_dir: Path | None = None,
    pixel_width: int | None = None,
) -> tuple[Mesh, int]:
    """
    Try to compute the mesh on on the image.
    First upscale the image with a given upscale factor
    If that yields only the trivial mesh lines, try to compute the mesh on
    the original image instead.
    Returns the mesh line coordinates and the scale factor used
    """
    upscaled_img = utils.scale_img(img, upscale_factor)
    mesh_lines = compute_mesh(
        upscaled_img, output_dir=output_dir, pixel_width=pixel_width
    )
    if not _is_trivial_mesh(mesh_lines):
        return mesh_lines, upscale_factor

    # If no mesh is found, then use the original image instead.
    fallback_mesh_lines = compute_mesh(
        img, output_dir=output_dir, pixel_width=pixel_width
    )
    return fallback_mesh_lines, 1


def _is_trivial_mesh(img_mesh: Mesh) -> bool:
    """
    Returns True if no lines have been identified when computing the mesh.
    That is, the points in mesh_x and mesh_y conist of the left, right, and top, bottom
    of the image respectively.
    """
    x_num = len(img_mesh[0])
    y_num = len(img_mesh[1])
    return x_num in (2, 3) and y_num in (2, 3)



"""Command line interface"""

import argparse
from pathlib import Path

from PIL import Image

from proper_pixel_art import pixelate


def add_pixelation_args(
    parser: argparse.ArgumentParser, group_name: str = "Pixelation options"
) -> argparse.ArgumentParser:
    """Add common pixelation arguments to an argument parser.

    Args:
        parser: The argument parser to add arguments to
        group_name: Name of the argument group (default: "Pixelation options")

    Returns:
        The parser with pixelation arguments added
    """
    pixel_group = parser.add_argument_group(group_name)
    pixel_group.add_argument(
        "-c",
        "--colors",
        dest="num_colors",
        type=int,
        default=None,
        help="Number of colors to quantize the image to (1-256). Omit to skip quantization and preserve all colors.",
    )
    pixel_group.add_argument(
        "-s",
        "--scale-result",
        dest="scale_result",
        type=int,
        default=1,
        help="Width of the 'pixels' in the output image (default: 1).",
    )
    pixel_group.add_argument(
        "-t",
        "--transparent",
        dest="transparent",
        action="store_true",
        default=False,
        help="Produce a transparent background in the output if set.",
    )
    pixel_group.add_argument(
        "-w",
        "--pixel-width",
        dest="pixel_width",
        type=int,
        default=None,
        help="Width of the pixels in the input image. If not set, it will be determined automatically.",
    )
    pixel_group.add_argument(
        "-u",
        "--initial-upscale",
        dest="initial_upscale",
        type=int,
        default=2,
        help=(
            "Initial image upscale factor in mesh detection algorithm. "
            "If the detected spacing is too large, "
            "it may be useful to increase this value."
        ),
    )
    return parser


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Generate a true-resolution pixel-art image from a source image."
    )
    parser.add_argument(
        "input_path", type=Path, nargs="?", help="Path to the source input file."
    )
    parser.add_argument(
        "-i",
        "--input",
        dest="input_path_flag",
        type=Path,
        help="Path to the source input file.",
    )
    parser.add_argument(
        "-o",
        "--output",
        dest="out_path",
        type=Path,
        default=Path("."),
        help="Path where the pixelated image will be saved. Can be either a directory or a file path.",
    )

    # Add common pixelation arguments
    add_pixelation_args(parser)

    args = parser.parse_args()

    # Either take the input as the first argument or use the -i flag
    if args.input_path is None and args.input_path_flag is None:
        parser.error("You must provide an input path (positional or with -i).")
    args.input_path = (
        args.input_path if args.input_path is not None else args.input_path_flag
    )

    return args


def resolve_output_path(
    out_path: Path, input_path: Path, suffix: str = "_pixelated"
) -> Path:
    """
    If outpath is a directory, make it a file path
    with filename e.g. (input stem)_pixelated.png
    """
    if out_path.suffix:
        return out_path
    filename = f"{input_path.stem}{suffix}.png"
    return out_path / filename


def main() -> None:
    args = parse_args()
    input_path = Path(args.input_path).expanduser()

    out_path = resolve_output_path(Path(args.out_path), input_path)
    out_path.parent.mkdir(exist_ok=True, parents=True)

    img = Image.open(input_path)
    pixelated = pixelate.pixelate(
        img,
        num_colors=args.num_colors,
        scale_result=args.scale_result,
        transparent_background=args.transparent,
        pixel_width=args.pixel_width,
        initial_upscale_factor=args.initial_upscale,
    )

    pixelated.save(out_path)


if __name__ == "__main__":
    main()



from pathlib import Path

from PIL import Image

from proper_pixel_art import mesh


def test_mesh():
    """
    Checks that the mesh calculated for the blob image is non-trivial.
    """
    img_path = Path.cwd() / "assets" / "blob" / "blob.png"
    img = Image.open(img_path).convert("RGBA")
    mesh_x, mesh_y = mesh.compute_mesh(img)
    assert (len(mesh_x)) > 2
    assert (len(mesh_y)) > 2



"""Main functions for pixelating an image with the pixelate function"""

from itertools import product
from pathlib import Path

import numpy as np
from PIL import Image

from proper_pixel_art import colors, mesh, utils
from proper_pixel_art.utils import Mesh


def downsample(
    image: Image.Image,
    mesh_lines: Mesh,
    skip_quantization: bool = False,
    original_alpha: np.ndarray | None = None,
) -> Image.Image:
    """
    Downsample the image by looping over each cell in mesh and
    selecting a representative color for each cell.

    Transparency handling:
    - If >=50% of pixels in a cell are transparent, the entire cell becomes transparent (0,0,0,0)
    - For skip_quantization=True, uses alpha channel from image directly
    - For skip_quantization=False with original_alpha provided,
      uses alpha channel from original_alpha array

    Args:
        image: The image to downsample (RGB if quantized, RGBA if not)
        mesh_lines: Tuple of (x_lines, y_lines) defining the pixel grid
        skip_quantization: If True, use histogram-based color selection for RGBA
        original_alpha: Optional numpy array of alpha channel values from original
                       image. Used to preserve transparency through quantization.
                       Only used when skip_quantization=False.

    Returns:
        RGBA image with downsampled pixels
    """
    lines_x, lines_y = mesh_lines
    height_result, width_result = len(lines_y) - 1, len(lines_x) - 1

    # Single conversion at start
    if skip_quantization:
        img_array = np.array(image.convert("RGBA"))
    else:
        img_array = np.array(image.convert("RGB"))

    # Output is RGBA to support transparency
    out = np.zeros((height_result, width_result, 4), dtype=np.uint8)

    for j, i in product(range(height_result), range(width_result)):
        x0, x1 = lines_x[i], lines_x[i + 1]
        y0, y1 = lines_y[j], lines_y[j + 1]
        cell = img_array[y0:y1, x0:x1]

        if skip_quantization:
            # Skip quantization and use original colros
            out[j, i] = colors.get_cell_color_skip_quantization(cell)
        else:
            # Get color from quantized cell, considering alpha from original
            if original_alpha is not None:
                cell_alpha = original_alpha[y0:y1, x0:x1]
                out[j, i] = colors.get_cell_color_with_alpha(cell, cell_alpha)
            else:
                out[j, i] = colors.get_opaque_cell_color(cell)

    return Image.fromarray(out, mode="RGBA")


def pixelate(
    image: Image.Image,
    num_colors: int | None = None,
    initial_upscale_factor: int = 2,
    scale_result: int | None = None,
    transparent_background: bool = False,
    intermediate_dir: Path | None = None,
    pixel_width: int | None = None,
) -> Image.Image:
    """
    Computes the true resolution pixel art image.
    inputs:
    - image:
        A PIL image to pixelate.
    - num_colors:
        The number of colors to use when quantizing the image.
        Use None to skip quantization and preserve all colors.
        This is an important parameter to tune,
        if it is too high, pixels that should be the same color will be different colors
        if it is too low, pixels that should be different colors will be the same color
    - scale_result:
        Upsample result by scale_result factor after algorithm is complete if not None.
    - initial_upscale_factor:
        Upsample original image by this factor. It may help detect lines.
    - transparent_background:
        If True, makes pixels matching the most common boundary color transparent.
        Applied after preserving original image transparency.
    - intermediate_dir:
        directory to save images visualizing intermediate steps.
    - pixel_width:
        If set, skips the step to automatically identify pixel width and uses this value.

    Returns the true pixelated image.
    """
    image_rgba = image.convert("RGBA")

    # Calculate the pixel mesh lines
    mesh_lines, upscale_factor = mesh.compute_mesh_with_scaling(
        image_rgba,
        initial_upscale_factor,
        output_dir=intermediate_dir,
        pixel_width=pixel_width,
    )

    # Process colors: either quantize or preserve original (with alpha)
    skip_quantization = num_colors is None
    if skip_quantization:
        # Preserve alpha: pass RGBA directly, let downsample filter by alpha
        processed_img = image_rgba
    else:
        processed_img = colors.palette_img(
            image_rgba, num_colors=num_colors, output_dir=intermediate_dir
        )

    # Scale the processed image to match the dimensions for the calculated mesh
    scaled_img = utils.scale_img(processed_img, upscale_factor)

    # Extract and scale alpha channel for quantized path
    scaled_alpha_array = (
        None
        if skip_quantization
        else colors.extract_and_scale_alpha(image_rgba, upscale_factor)
    )

    # Downsample the image to 1 pixel per cell in the mesh
    result = downsample(
        scaled_img,
        mesh_lines,
        skip_quantization=skip_quantization,
        original_alpha=scaled_alpha_array,
    )

    if transparent_background:
        result = colors.make_background_transparent(result)

    if scale_result is not None:
        result = utils.scale_img(result, int(scale_result))

    return result



"""Web interface for Proper Pixel Art using Gradio."""

from PIL import Image

from proper_pixel_art.pixelate import pixelate

IMG_HEIGHT = 512


def process(
    image: Image.Image | None,
    num_colors: int,
    transparent: bool,
    scale: int,
    initial_upscale: int,
    pixel_width: int,
) -> Image.Image | None:
    """Process image through pixelation pipeline."""
    if image is None:
        return None
    return pixelate(
        image,
        num_colors=num_colors if num_colors > 0 else None,
        transparent_background=transparent,
        scale_result=scale if scale > 1 else None,
        initial_upscale_factor=initial_upscale,
        pixel_width=pixel_width if pixel_width > 0 else None,
    )


def create_demo():
    """Create Gradio demo interface."""
    import gradio as gr

    with gr.Blocks(title="Proper Pixel Art") as demo:
        gr.Markdown(
            "# Proper Pixel Art\nConvert AI-generated pixel art to true pixel resolution"
        )

        with gr.Row():
            with gr.Column():
                input_img = gr.Image(
                    type="pil",
                    label="Input",
                    format="png",
                    image_mode="RGBA",
                    height=IMG_HEIGHT,
                )
            with gr.Column():
                output_img = gr.Image(
                    type="pil",
                    label="Output",
                    format="png",
                    image_mode="RGBA",
                    height=IMG_HEIGHT,
                    interactive=False,
                )

        with gr.Row():
            num_colors = gr.Slider(
                0, 64, value=16, step=1, label="Colors (0 = skip quantization)"
            )
            scale = gr.Slider(1, 20, value=1, step=1, label="Scale Result")

        with gr.Row():
            initial_upscale = gr.Slider(1, 4, value=2, step=1, label="Initial Upscale")
            pixel_width = gr.Slider(
                0, 50, value=0, step=1, label="Pixel Width (0=auto)"
            )

        with gr.Row():
            transparent = gr.Checkbox(value=False, label="Transparent Background")
            btn = gr.Button("Pixelate", variant="primary")

        btn.click(
            fn=process,
            inputs=[
                input_img,
                num_colors,
                transparent,
                scale,
                initial_upscale,
                pixel_width,
            ],
            outputs=output_img,
        )

    return demo


def main():
    """Entry point for ppa-web command."""
    demo = create_demo()
    demo.launch()


if __name__ == "__main__":
    main()
